/*
 * Copyright (c) 2009 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 *
 * Author(s): Torsten Hoefler <htor@cs.indiana.edu>
 *
 */
#include "nbcbench.h"

#ifdef SYNC_BARRIER
/* simple MPI_Barrier synchronization mechanism */
void sync_init_stage1(t_mpiargs *mpiargs) {

}

void sync_init_stage2(t_mpiargs *mpiargs) {

}

double nbcb_sync(t_mpiargs *mpiargs) {
  MPI_Barrier(mpiargs->comm);
  return 0;
}
#endif

#ifdef SYNC_DISSEMINATION
/* internal dissemination algorithm */
void sync_init_stage1(t_mpiargs *mpiargs) {

}

void sync_init_stage2(t_mpiargs *mpiargs) {

}

/* done at the beginning of every new nprocs */
double nbcb_sync(t_mpiargs *mpiargs) {
  MPI_Comm comm = mpiargs->comm;
  int round = -1, src, dst, maxround, size, i, rank;
  char buf;

  MPI_Comm_size(comm, &size);
  MPI_Comm_rank(comm, &rank);

  /* round starts with 0 -> round - 1 */
  maxround = (int)ceil((double)log(size)) - 1;
            
  do {
    round++;
    dst = (rank + round) % size;
    MPI_Send(&buf, 1, MPI_BYTE, dst, 0, comm);
    /* + size*size that the modulo does never get negative ! */
    src = ((rank - round) + size*size) % size;
    MPI_Recv(&buf, 1, MPI_BYTE, src, 0, comm, MPI_STATUS_IGNORE);
  } while (round < maxround);     

  return 0;
}
#endif

#ifdef SYNC_WINDOW
static double *diffs=NULL; /* global array of all diffs to all ranks - only
                  completely valid on rank 0 */
static double gdiff=0; /* the is the final time diff to rank 0 :-) */
static double gnext=0; /* start-time for next round - synchronized with rank 0 */

#define NUMBER_SMALLER 100 /* do RTT measurement until n successive
                             messages are *not* smaller than the current
                             smallest one */

/* time-window based synchronization mechanism 
 * - */
void sync_init_stage1(t_mpiargs *mpiargs) {
  int p, r, res, dist, round;


  res = MPI_Comm_rank(mpiargs->comm, &r);
  res = MPI_Comm_size(mpiargs->comm, &p);
  
  /* reallocate tha diffs array with the right size */
  if(diffs != NULL) free(diffs);
  diffs = (double*)calloc(1, p*sizeof(double));
  
  /* check if p is power of 2 */
  { int i=1;
    while((i = i << 1) < p) {};
    if(i != p) { 
      printf("communicator size (%i) must be power of 2 (%i)!\n", p, i);
      MPI_Abort(MPI_COMM_WORLD, 1);
    }
  }

  dist = 1; /* this gets left-shifted (<<) every round and is after 
                   $\lceil log_2(p) \rceil$ rounds >= p */
  round = 1; /* fun and printf round counter - not really needed */
  do {
    int peer; /* synchronization peer */
    int client, server;
    double tstart, /* local start time */
           tend, /* local end time */
           trem, /* remote time */
           tmpdiff, /* temporary difference to remote clock */
           diff; /* difference to remote clock */
    client = 0; server = 0;
    client = ((r % (dist << 1)) == 0);
    server = ((r % (dist << 1)) == dist);
    
    if(server) {
      peer = r - dist;
      if(peer < 0) server = 0;
      //if(server) printf("(%i) %i <- %i\n", round, r, peer);
    }
    if(client) {
      peer = r + dist;
      if(peer >= p) client = 0;
      //if(client) printf("(%i) %i -> %i\n", round, peer, r);
    }
    if(!client && !server) break; /* TODO: leave loop if no peer left -
                                     works only for power of two process
                                     groups */

    /* synchronize clocks with peer */
    {
    int notsmaller = 0; /* count number of RTTs that are *not* smaller than
                    the current smallest one */
    double smallest = MAX_DOUBLE; /* the current smallest time */
    do {
      /* the client sends a ping to the server and waits for a pong (and
       * takes the RTT time). It repeats this procedure until the last
       * NUMBER_SMALLER RTTs have not been smaller than the smallest
       * (tries to find the smallest RTT). When the smallest RTT is
       * found, it sends a special flag (0d) to the server that it knows
       * that the benchmark is finished. The client computes the diff
       * with this smallest RTT with the scheme described in the paper.
       * */
      if(client) {
        tstart = MPI_Wtime();
        res = MPI_Send(&tstart, 1, MPI_DOUBLE, peer, 0, mpiargs->comm);
        //printf("[%i] client: waiting for response from %i\n", r, peer);
        res = MPI_Recv(&trem, 1, MPI_DOUBLE, peer, 0, mpiargs->comm, MPI_STATUS_IGNORE);
        //printf("[%i] client: got response from %i\n", r, peer);
        tend = MPI_Wtime();
        tmpdiff = tstart + (tend-tstart)/2 - trem;
        
        if(tend-tstart < smallest) {
          smallest = tend-tstart;
          notsmaller = 0;
          diff = tmpdiff; /* save new smallest diff-time */
        } else {
          if(++notsmaller == NUMBER_SMALLER) {
            /* send abort flag to client */
            trem = 0;
            res = MPI_Send(&trem, 1, MPI_DOUBLE, peer, 0, mpiargs->comm);
            //printf("[%i] diff to %i: %lf\n", r, peer, diff*1e6);
            break;
          }
        }
        //printf("[%i] notsmaller: %i\n", r, notsmaller);
      }

      /* The server just replies with the local time to the client
       * requests and aborts the benchmark if the abort flag (0d) is
       * received in any of the requests. */
      if(server) {
        //printf("[%i] server: waiting for ping from %i\n", r, peer);
        res = MPI_Recv(&tstart, 1, MPI_DOUBLE, peer, 0, mpiargs->comm, MPI_STATUS_IGNORE);
        if(tstart == 0) break; /* this is the signal from the client to stop */
        trem = MPI_Wtime(); /* fill in local time on server */
        //printf("[%i] server: got ping from %i (%lf) \n", r, peer, tstart);
        res = MPI_Send(&trem, 1, MPI_DOUBLE, peer, 0, mpiargs->comm);
      }
      /* this loop is only left with a break */
    } while(1);
    }
    
    /* the client measured the time difference to his peer-server of the
     * current round. Since rank 0 is the global synchronization point,
     * rank 0's array has to be up to date and the other clients have to
     * communicate all their knowledge to rank 0 as described in the
     * paper. */
    
    if(client) {
      /* all clients just measured the time difference to node r + diff
       * (=peer) */
      diffs[peer] = diff;

      /* we are a client - we need to receive all the knowledge
       * (differences) that the server we just synchronized with holds!
       * Our server has been "round-1" times client and measures
       * "round-1" diffs */
      if(round > 1) {
        double *recvbuf; /* receive the server's data */
        int items, i;

        items = (1 << (round-1))-1;
        recvbuf = (double*)malloc(items*sizeof(double));
        
        res = MPI_Recv(recvbuf, items, MPI_DOUBLE, peer, 0, mpiargs->comm, MPI_STATUS_IGNORE);
        
        //printf("[%i] round: %i, client merges %i items\n", r, round, items);
        /* merge data into my own field */
        for(i=0; i<items; i++) {
          diffs[peer+i+1] =  diffs[peer] /* diff to server */ + 
                          recvbuf[i] /* received time */; // TODO: + or - ???
        }
        free(recvbuf);
      }
    }

    if(server) {
      /* we are a server, we need to send all our knowledge (time
       * differences to our client */
    
      /* we have measured "round-1" nodes at the end of round "round"
       * and hold $2^(round-1)-1$ diffs at this time*/
      if(round > 1) {
        int i, tmpdist, tmppeer, items;
        double *sendbuf;
        
        items = (1 << (round-1))-1;
        sendbuf = (double*)malloc(items*sizeof(double));

        //printf("[%i] round: %i, server sends %i items\n", r, round, items);

        /* fill buffer - every server holds the $2^(round-1)-1$ next
         * diffs */
        for(i=0; i<items; i++) {
          sendbuf[i] = diffs[r+i+1];
        }
        res = MPI_Send(sendbuf, items, MPI_DOUBLE, peer, 0, mpiargs->comm);
        free(sendbuf);
      }
    }
    
    dist = dist << 1;
    round++;
  } while(dist < p);

  /*if(!r) {
    int i;
    for(i=0; i<p; i++) {
      printf("%lf ", diffs[i]*1e6);
    }
    printf("\n");
  }*/

  /* scatter all the time diffs to the processes */
  MPI_Scatter(diffs, 1, MPI_DOUBLE, &gdiff, 1, MPI_DOUBLE, 0, mpiargs->comm);
  //printf("[%i] diff: %lf\n", r, gdiff*1e6);

  /* initialize window to 0 */
  mpiargs->window = 0;
}

/* done at the beginning of every new size */
void sync_init_stage2(t_mpiargs *mpiargs) {
  double bcasttime, win; /* measure MPI_Bcast() time :-/ */
  int i;

#if 0
  /* we increased the window size? - we need to re-synch */
  if(mpiargs->window != 0) {
    /* save increased window size */
    win = mpiargs->window;
    sync_init_stage1(mpiargs);
    mpiargs->window = win;
  }
#endif

  //if(!mpiargs->rank) printf("estimated time: %lf\n", mpiargs->esttime);
  win = mpiargs->esttime/1e6 * 1.25; /* esttime is in usec - window
                                            length = 125% of the
                                            estimated MPI time */

  if(win > mpiargs->window) mpiargs->window=win;
  MPI_Bcast(&mpiargs->window, 1, MPI_DOUBLE, 0, mpiargs->comm);
  
  /* it has been used before - so it is "warm" :) */
  bcasttime = -MPI_Wtime();
  for(i=0; i<10; i++) {
    MPI_Bcast(&gnext, 1, MPI_DOUBLE, 0, mpiargs->comm);
  }
  bcasttime += MPI_Wtime();
  gnext = bcasttime; /* dummy buffer */
  /* get maximum bcasttime */
  MPI_Reduce(&gnext, &bcasttime, 1, MPI_DOUBLE, MPI_MAX, 0, mpiargs->comm);

  /* rank 0 sets base-time to a time when the bcast is expected to be
   * finished */
  gnext = MPI_Wtime() + bcasttime;
  MPI_Bcast(&gnext, 1, MPI_DOUBLE, 0, mpiargs->comm);

  gnext= gnext - gdiff; /* adjust rank 0's time to local time */
  //if(!mpiargs->rank) printf("[%i] gnext: %lf, bcasttime: %lf, window: %lf\n", mpiargs->rank, gnext*1e6, 1e6*bcasttime, mpiargs->window*1e6);
}

int syncctr=0;
static volatile int NBC_Dummy_var=0; /* avoid optimizations */

/* for every single measurement */
double nbcb_sync(t_mpiargs *mpiargs) {
  double err = 0;
 
  /* OMPI does not send messages immediately!!!! -> drain
   * messages */
  MPI_Barrier(mpiargs->comm);

  if(MPI_Wtime() > gnext) {
    //printf("[%i] tooooo late!!!! :-( (%lf)\n", mpiargs->rank, (MPI_Wtime()-gnext)*1e6);
    err = MPI_Wtime()-gnext;
  } else {
    /* wait */
    //printf("waiting ... %lf\n", (gnext-MPI_Wtime()));
    while(MPI_Wtime() < gnext) {NBC_Dummy_var++;};
  }
  
  //printf("[%i] syncctr: %i (next: %lf, err: %lf)\n", mpiargs->rank, syncctr++, gnext*1e6, err*1e6);
  
  gnext = gnext + mpiargs->window;
  //printf("leaving ... err: %lf, winlen: %lf\n", err, mpiargs->window*1e6);
  return err;
}
#endif
